@@ -8,7 +8,9 @@
 # See https://github.com/nexB/skeleton for support or download.
 # See https://aboutcode.org for more information about nexB OSS projects.
 #
+import asyncio
 import email
+import hashlib
 import itertools
 import os
 import pathlib
@@ -17,13 +19,18 @@ import shutil
 import tempfile
 import time
 from collections import defaultdict
+from typing import Dict
 from typing import List
 from typing import NamedTuple
+from typing import Tuple
+from typing import Union
 from urllib.parse import quote_plus
 from urllib.parse import unquote
 from urllib.parse import urlparse
 from urllib.parse import urlunparse
 
+import aiofiles
+import aiohttp
 import attr
 import packageurl
 import requests
@@ -34,7 +41,8 @@ from packvers import tags as packaging_tags
 from packvers import version as packaging_version
 from packvers.specifiers import SpecifierSet
 
-from python_inspector import DEFAULT_PYTHON_VERSION
+from python_inspector import lockfile
+from python_inspector import pyinspector_settings as settings
 from python_inspector import utils_pip_compatibility_tags
 
 """
@@ -145,6 +153,8 @@ PLATFORMS_BY_OS = {
         "manylinux1_x86_64",
         "manylinux2010_x86_64",
         "manylinux2014_x86_64",
+        "manylinux_2_27_x86_64",
+        "manylinux_2_28_x86_64",
         "manylinux2014_aarch6",
         "musllinux_1_2_x86_64",
         "manylinux_2_33_aarch64",
@@ -185,21 +195,8 @@ PLATFORMS_BY_OS = {
     ],
 }
 
-CACHE_THIRDPARTY_DIR = os.environ.get("PYTHON_INSPECTOR_CACHE_DIR")
-if not CACHE_THIRDPARTY_DIR:
-    CACHE_THIRDPARTY_DIR = ".cache/python_inspector"
-    try:
-        os.makedirs(CACHE_THIRDPARTY_DIR, exist_ok=True)
-    except Exception:
-        home = pathlib.Path.home()
-        CACHE_THIRDPARTY_DIR = str(home / ".cache/python_inspector")
-        os.makedirs(CACHE_THIRDPARTY_DIR, exist_ok=True)
-
-
-################################################################################
-
-PYPI_SIMPLE_URL = "https://pypi.org/simple"
-PYPI_INDEX_URLS = (PYPI_SIMPLE_URL,)
+DEFAULT_PYTHON_VERSION = settings.DEFAULT_PYTHON_VERSION
+CACHE_THIRDPARTY_DIR = settings.CACHE_THIRDPARTY_DIR
 
 ################################################################################
 
@@ -221,7 +218,7 @@ class DistributionNotFound(Exception):
     pass
 
 
-def download_wheel(
+async def download_wheel(
     name,
     version,
     environment,
@@ -243,11 +240,11 @@ def download_wheel(
         print(f"  download_wheel: {name}=={version} for envt: {environment}")
 
     if not repos:
-        repos = DEFAULT_PYPI_REPOS
+        raise ValueError("download_wheel: missing repos")
 
     fetched_wheel_filenames = []
     for repo in repos:
-        supported_and_valid_wheels = get_supported_and_valid_wheels(
+        supported_and_valid_wheels = await get_supported_and_valid_wheels(
             repo, name, version, environment, python_version
         )
         if not supported_and_valid_wheels:
@@ -257,7 +254,8 @@ def download_wheel(
                 )
             continue
         for wheel in supported_and_valid_wheels:
-            fetched_wheel_filename = wheel.download(
+            wheel.credentials = repo.credentials
+            fetched_wheel_filename = await wheel.download(
                 dest_dir=dest_dir,
                 verbose=verbose,
                 echo_func=echo_func,
@@ -265,13 +263,13 @@ def download_wheel(
             fetched_wheel_filenames.append(fetched_wheel_filename)
 
         if fetched_wheel_filenames:
-            # do not futher fetch from other repos if we find in first, typically PyPI
+            # do not further fetch from other repos if we find in first, typically PyPI
             break
     return fetched_wheel_filenames
 
 
-def get_valid_sdist(repo, name, version, python_version=DEFAULT_PYTHON_VERSION):
-    package = repo.get_package_version(name=name, version=version)
+async def get_valid_sdist(repo, name, version, python_version=DEFAULT_PYTHON_VERSION):
+    package = await repo.get_package_version(name=name, version=version)
     if not package:
         if TRACE_DEEP:
             print(
@@ -292,13 +290,13 @@ def get_valid_sdist(repo, name, version, python_version=DEFAULT_PYTHON_VERSION):
     return sdist
 
 
-def get_supported_and_valid_wheels(
+async def get_supported_and_valid_wheels(
     repo, name, version, environment, python_version=DEFAULT_PYTHON_VERSION
 ) -> List:
     """
     Return a list of wheels matching the ``environment`` Environment constraints.
     """
-    package = repo.get_package_version(name=name, version=version)
+    package = await repo.get_package_version(name=name, version=version)
     if not package:
         if TRACE_DEEP:
             print(
@@ -319,9 +317,10 @@ def get_supported_and_valid_wheels(
         ):
             continue
         if TRACE_DEEP:
+            durl = await wheel.download_url(repo)
             print(
                 f"""    get_supported_and_valid_wheels: Getting wheel from index (or cache):
-                {wheel.download_url}"""
+                {durl}"""
             )
         wheels.append(wheel)
     return wheels
@@ -336,7 +335,7 @@ def valid_python_version(python_version, python_requires):
     return python_version in SpecifierSet(python_requires)
 
 
-def download_sdist(
+async def download_sdist(
     name,
     version,
     dest_dir=CACHE_THIRDPARTY_DIR,
@@ -356,24 +355,24 @@ def download_sdist(
         print(f"  download_sdist: {name}=={version}")
 
     if not repos:
-        repos = DEFAULT_PYPI_REPOS
+        raise ValueError("download_sdist: missing repos")
 
     fetched_sdist_filename = None
 
     for repo in repos:
-        sdist = get_valid_sdist(repo, name, version, python_version=python_version)
+        sdist = await get_valid_sdist(repo, name, version, python_version=python_version)
         if not sdist:
             if TRACE_DEEP:
                 print(f"    download_sdist: No valid sdist for {name}=={version}")
             continue
-        fetched_sdist_filename = sdist.download(
+        fetched_sdist_filename = await sdist.download(
             dest_dir=dest_dir,
             verbose=verbose,
             echo_func=echo_func,
         )
 
         if fetched_sdist_filename:
-            # do not futher fetch from other repos if we find in first, typically PyPI
+            # do not further fetch from other repos if we find in first, typically PyPI
             break
 
     return fetched_sdist_filename
@@ -433,7 +432,6 @@ class Link(NamedTuple):
 
 @attr.attributes
 class Distribution(NameVer):
-
     """
     A Distribution is either either a Wheel or Sdist and is identified by and
     created from its filename as well as its name and version. A Distribution is
@@ -611,11 +609,12 @@ class Distribution(NameVer):
             )
         )
 
-    @property
-    def download_url(self):
-        return self.get_best_download_url()
+    async def download_url(self, repo):
+        if not repo:
+            raise ValueError("download_url: missing repo")
+        return await self.get_best_download_url(repos=(repo,))
 
-    def get_best_download_url(self, repos=tuple()):
+    async def get_best_download_url(self, repos=tuple()):
         """
         Return the best download URL for this distribution where best means this
         is the first URL found for this distribution found in the list of
@@ -625,10 +624,10 @@ class Distribution(NameVer):
         """
 
         if not repos:
-            repos = DEFAULT_PYPI_REPOS
+            raise ValueError("get_best_download_url: missing repos")
 
         for repo in repos:
-            package = repo.get_package_version(name=self.name, version=self.version)
+            package = await repo.get_package_version(name=self.name, version=self.version)
             if not package:
                 if TRACE:
                     print(
@@ -645,7 +644,7 @@ class Distribution(NameVer):
                         f"     get_best_download_url: {self.filename} not found in {repo.index_url}"
                     )
 
-    def download(
+    async def download(
         self,
         dest_dir=CACHE_THIRDPARTY_DIR,
         verbose=False,
@@ -663,7 +662,7 @@ class Distribution(NameVer):
             )
 
         # FIXME:
-        fetch_and_save(
+        await fetch_and_save(
             path_or_url=self.path_or_url,
             dest_dir=dest_dir,
             credentials=self.credentials,
@@ -1005,7 +1004,6 @@ class Sdist(Distribution):
 
 @attr.attributes
 class Wheel(Distribution):
-
     """
     Represents a wheel file.
 
@@ -1268,12 +1266,12 @@ class PypiPackage(NameVer):
         return package
 
     @classmethod
-    def packages_from_links(cls, links: List[Link]):
+    async def packages_from_links(cls, links: List[Link]):
         """
         Yield PypiPackages built from a list of paths or URLs.
         These are sorted by name and then by version from oldest to newest.
         """
-        dists = PypiPackage.dists_from_links(links)
+        dists = await PypiPackage.dists_from_links(links)
         if TRACE_ULTRA_DEEP:
             print("packages_from_many_paths_or_urls: dists:", dists)
 
@@ -1289,7 +1287,7 @@ class PypiPackage(NameVer):
             yield package
 
     @classmethod
-    def dists_from_links(cls, links: List[Link]):
+    async def dists_from_links(cls, links: List[Link]):
         """
         Return a list of Distribution given a list of
         ``paths_or_urls`` to wheels or source distributions.
@@ -1307,7 +1305,7 @@ class PypiPackage(NameVer):
         ...     Link(url="https://example.com/bar/bitarray-0.8.1.tar.gz",python_requires= ">=3.7"),
         ...     Link(url="bitarray-0.8.1.tar.gz.ABOUT",python_requires= ">=3.7"),
         ...     Link(url="bit.LICENSE", python_requires=">=3.7")]
-        >>> results = list(PypiPackage.dists_from_links(links))
+        >>> results = list(asyncio.run(PypiPackage.dists_from_links(links)))
         >>> for r in results:
         ...    print(r.__class__.__name__, r.name, r.version)
         ...    if isinstance(r, Wheel):
@@ -1333,9 +1331,6 @@ class PypiPackage(NameVer):
                         "     ===> dists_from_paths_or_urls:",
                         dist,
                         "\n     ",
-                        "with URL:",
-                        dist.download_url,
-                        "\n     ",
                         "from URL:",
                         link.url,
                     )
@@ -1469,7 +1464,8 @@ class PypiSimpleRepository:
 
     index_url = attr.ib(
         type=str,
-        default=PYPI_SIMPLE_URL,
+        # We use first entry in INDEX_URL setting as default
+        default=settings.INDEX_URL[0],
         metadata=dict(help="Base PyPI simple URL for this index."),
     )
 
@@ -1504,7 +1500,7 @@ class PypiSimpleRepository:
 
     credentials = attr.ib(type=dict, default=None)
 
-    def _get_package_versions_map(
+    async def _get_package_versions_map(
         self,
         name,
         verbose=False,
@@ -1520,7 +1516,7 @@ class PypiSimpleRepository:
         if not versions and normalized_name not in self.fetched_package_normalized_names:
             self.fetched_package_normalized_names.add(normalized_name)
             try:
-                links = self.fetch_links(
+                links = await self.fetch_links(
                     normalized_name=normalized_name,
                     verbose=verbose,
                     echo_func=echo_func,
@@ -1528,7 +1524,7 @@ class PypiSimpleRepository:
                 # note that this is sorted so the mapping is also sorted
                 versions = {
                     package.version: package
-                    for package in PypiPackage.packages_from_links(links=links)
+                    async for package in PypiPackage.packages_from_links(links=links)
                 }
                 self.packages[normalized_name] = versions
             except RemoteNotFetchedException as e:
@@ -1540,26 +1536,26 @@ class PypiSimpleRepository:
 
         return versions
 
-    def get_package_versions(
+    async def get_package_versions(
         self,
         name,
         verbose=False,
         echo_func=None,
-    ):
+    ) -> Dict:
         """
         Return a mapping of all available PypiPackage version as{version:
         package} for this package name. The mapping may be empty but not None.
         It is sorted by version from oldest to newest.
         """
         return dict(
-            self._get_package_versions_map(
+            await self._get_package_versions_map(
                 name=name,
                 verbose=verbose,
                 echo_func=echo_func,
             )
         )
 
-    def get_package_version(
+    async def get_package_version(
         self,
         name,
         version=None,
@@ -1572,22 +1568,26 @@ class PypiSimpleRepository:
         """
         if not version:
             versions = list(
-                self._get_package_versions_map(
-                    name=name,
-                    verbose=verbose,
-                    echo_func=echo_func,
+                (
+                    await self._get_package_versions_map(
+                        name=name,
+                        verbose=verbose,
+                        echo_func=echo_func,
+                    )
                 ).values()
             )
             # return the latest version
             return versions and versions[-1]
         else:
-            return self._get_package_versions_map(
-                name=name,
-                verbose=verbose,
-                echo_func=echo_func,
+            return (
+                await self._get_package_versions_map(
+                    name=name,
+                    verbose=verbose,
+                    echo_func=echo_func,
+                )
             ).get(version)
 
-    def fetch_links(
+    async def fetch_links(
         self,
         normalized_name,
         verbose=False,
@@ -1598,7 +1598,7 @@ class PypiSimpleRepository:
         name using the `index_url` of this repository.
         """
         package_url = f"{self.index_url}/{normalized_name}"
-        text = CACHE.get(
+        text, _ = await CACHE.get(
             path_or_url=package_url,
             credentials=self.credentials,
             as_text=True,
@@ -1645,16 +1645,14 @@ def resolve_relative_url(package_url, url):
     return url
 
 
-PYPI_PUBLIC_REPO = PypiSimpleRepository(index_url=PYPI_SIMPLE_URL)
-DEFAULT_PYPI_REPOS = (PYPI_PUBLIC_REPO,)
-DEFAULT_PYPI_REPOS_BY_URL = {r.index_url: r for r in DEFAULT_PYPI_REPOS}
-
 ################################################################################
 #
 # Basic file and URL-based operations using a persistent file-based Cache
 #
 ################################################################################
 
+PYINSP_CACHE_LOCK_TIMEOUT = 120  # in seconds
+
 
 @attr.attributes
 class Cache:
@@ -1668,7 +1666,10 @@ class Cache:
     def __attrs_post_init__(self):
         os.makedirs(self.directory, exist_ok=True)
 
-    def get(
+    def sha256_hash(self, text: str) -> str:
+        return hashlib.sha256(text.encode()).hexdigest()
+
+    async def get(
         self,
         credentials,
         path_or_url,
@@ -1676,20 +1677,22 @@ class Cache:
         force=False,
         verbose=False,
         echo_func=None,
-    ):
+    ) -> Tuple[Union[str, bytes], str]:
         """
         Return the content fetched from a ``path_or_url`` through the cache.
         Raise an Exception on errors. Treats the content as text if as_text is
         True otherwise as treat as binary. `path_or_url` can be a path or a URL
         to a file.
         """
-        cache_key = quote_plus(path_or_url.strip("/"))
+        # the cache key is a hash of the normalized path
+        cache_key = self.sha256_hash(quote_plus(path_or_url.strip("/")))
         cached = os.path.join(self.directory, cache_key)
+        lock_file = f"{cached}.lockfile"
 
         if force or not os.path.exists(cached):
             if TRACE_DEEP:
                 print(f"        FILE CACHE MISS: {path_or_url}")
-            content = get_file_content(
+            content = await get_file_content(
                 path_or_url=path_or_url,
                 credentials=credentials,
                 as_text=as_text,
@@ -1697,19 +1700,24 @@ class Cache:
                 echo_func=echo_func,
             )
             wmode = "w" if as_text else "wb"
-            with open(cached, wmode) as fo:
-                fo.write(content)
-            return content
+
+            # acquire lock and wait until timeout to get a lock or die
+            with lockfile.FileLock(lock_file).locked(timeout=PYINSP_CACHE_LOCK_TIMEOUT):
+                async with aiofiles.open(cached, mode=wmode) as fo:
+                    await fo.write(content)
+            return content, cached
         else:
             if TRACE_DEEP:
                 print(f"        FILE CACHE HIT: {path_or_url}")
-            return get_local_file_content(path=cached, as_text=as_text)
+            # also lock on read to avoid race conditions
+            with lockfile.FileLock(lock_file).locked(timeout=PYINSP_CACHE_LOCK_TIMEOUT):
+                return await get_local_file_content(path=cached, as_text=as_text), cached
 
 
 CACHE = Cache()
 
 
-def get_file_content(
+async def get_file_content(
     path_or_url,
     credentials,
     as_text=True,
@@ -1723,7 +1731,7 @@ def get_file_content(
     if path_or_url.startswith("https://"):
         if TRACE_DEEP:
             print(f"Fetching: {path_or_url}")
-        _headers, content = get_remote_file_content(
+        _headers, content = await get_remote_file_content(
             url=path_or_url,
             credentials=credentials,
             as_text=as_text,
@@ -1735,13 +1743,13 @@ def get_file_content(
     elif path_or_url.startswith("file://") or (
         path_or_url.startswith("/") and os.path.exists(path_or_url)
     ):
-        return get_local_file_content(path=path_or_url, as_text=as_text)
+        return await get_local_file_content(path=path_or_url, as_text=as_text)
 
     else:
         raise Exception(f"Unsupported URL scheme: {path_or_url}")
 
 
-def get_local_file_content(path, as_text=True):
+async def get_local_file_content(path: str, as_text=True) -> str:
     """
     Return the content at `url` as text. Return the content as bytes is
     `as_text` is False.
@@ -1750,15 +1758,15 @@ def get_local_file_content(path, as_text=True):
         path = path[7:]
 
     mode = "r" if as_text else "rb"
-    with open(path, mode) as fo:
-        return fo.read()
+    async with aiofiles.open(path, mode=mode) as fo:
+        return await fo.read()
 
 
 class RemoteNotFetchedException(Exception):
     pass
 
 
-def get_remote_file_content(
+async def get_remote_file_content(
     url,
     credentials,
     as_text=True,
@@ -1770,7 +1778,7 @@ def get_remote_file_content(
 ):
     """
     Fetch and return a tuple of (headers, content) at `url`. Return content as a
-    text string if `as_text` is True. Otherwise return the content as bytes.
+    text string if `as_text` is True. Otherwise, return the content as bytes.
 
     If `header_only` is True, return only (headers, None). Headers is a mapping
     of HTTP headers.
@@ -1779,7 +1787,7 @@ def get_remote_file_content(
     """
     time.sleep(_delay)
     headers = headers or {}
-    # using a GET with stream=True ensure we get the the final header from
+    # using a GET with stream=True ensure we get the final header from
     # several redirects and that we can ignore content there. A HEAD request may
     # not get us this last header
     if verbose and not echo_func:
@@ -1791,39 +1799,32 @@ def get_remote_file_content(
     if credentials:
         auth = (credentials.get("login"), credentials.get("password"))
 
-    stream = requests.get(
-        url,
-        allow_redirects=True,
-        stream=True,
-        headers=headers,
-        auth=auth,
-    )
+    async with aiohttp.ClientSession() as session:
+        async with session.get(url, allow_redirects=True, headers=headers, auth=auth) as response:
+            status = response.status
+            if status != requests.codes.ok:  # NOQA
+                if status == 429 and _delay < 20:
+                    # too many requests: start some exponential delay
+                    increased_delay = (_delay * 2) or 1
 
-    with stream as response:
-        status = response.status_code
-        if status != requests.codes.ok:  # NOQA
-            if status == 429 and _delay < 20:
-                # too many requests: start some exponential delay
-                increased_delay = (_delay * 2) or 1
+                    return await get_remote_file_content(
+                        url,
+                        credentials=credentials,
+                        as_text=as_text,
+                        headers_only=headers_only,
+                        _delay=increased_delay,
+                    )
 
-                return get_remote_file_content(
-                    url,
-                    credentials=credentials,
-                    as_text=as_text,
-                    headers_only=headers_only,
-                    _delay=increased_delay,
-                )
+                else:
+                    raise RemoteNotFetchedException(f"Failed HTTP request from {url} with {status}")
 
-            else:
-                raise RemoteNotFetchedException(f"Failed HTTP request from {url} with {status}")
+            if headers_only:
+                return response.headers, None
 
-        if headers_only:
-            return response.headers, None
-
-        return response.headers, response.text if as_text else response.content
+            return response.headers, await response.text() if as_text else await response.read()
 
 
-def fetch_and_save(
+async def fetch_and_save(
     path_or_url,
     dest_dir,
     filename,
@@ -1834,19 +1835,28 @@ def fetch_and_save(
 ):
     """
     Fetch content at ``path_or_url`` URL or path and save this to
-    ``dest_dir/filername``. Return the fetched content. Raise an Exception on
+    ``dest_dir/filename``. Return the fetched content. Raise an Exception on
     errors. Treats the content as text if as_text is True otherwise as treat as
     binary.
     """
-    content = CACHE.get(
+    content, path = await CACHE.get(
         path_or_url=path_or_url,
         credentials=credentials,
         as_text=as_text,
         verbose=verbose,
         echo_func=echo_func,
     )
+
     output = os.path.join(dest_dir, filename)
-    wmode = "w" if as_text else "wb"
-    with open(output, wmode) as fo:
-        fo.write(content)
+    if os.path.exists(output):
+        os.remove(output)
+
+    os.symlink(os.path.abspath(path), output)
     return content
+
+
+def get_current_indexes() -> list[PypiSimpleRepository]:
+    """
+    Return a list of PypiSimpleRepository indexes configured in settings.
+    """
+    return [PypiSimpleRepository(index_url=url) for url in settings.INDEX_URL]
