@@ -12,6 +12,8 @@
 import json
 import os
 import sys
+import time
+from os.path import dirname
 
 import pytest
 from click.testing import CliRunner
@@ -28,6 +30,15 @@ setup_test_env = FileDrivenTesting()
 setup_test_env.test_data_dir = os.path.join(os.path.dirname(__file__), "data", "setup")
 
 
+def clear_environ():
+    for k, v in os.environ.items():
+        if k == "PYINSP_REGEN_TEST_FIXTURES":
+            continue
+        if "PYINSP" in k:
+            print(f"deleting: {k}: {v}")
+            os.unsetenv(k)
+
+
 @pytest.mark.online
 def test_cli_with_default_urls():
     expected_file = test_env.get_test_loc("default-url-expected.json", must_exist=False)
@@ -43,6 +54,17 @@ def test_cli_with_default_urls():
     )
 
 
+@pytest.mark.online
+def test_cli_with_specs_with_no_install_requires():
+    expected_file = test_env.get_test_loc("no-install-requires-expected.json", must_exist=False)
+    specifier = "crontab==1.0.4"
+    check_specs_resolution(
+        specifier=specifier,
+        expected_file=expected_file,
+        regen=REGEN_TEST_FIXTURES,
+    )
+
+
 @pytest.mark.online
 def test_cli_with_requirements_and_ignore_errors():
     requirements_file = test_env.get_test_loc("error-requirements.txt")
@@ -134,6 +156,8 @@ def test_cli_with_single_index_url_except_pypi_simple():
     extra_options = [
         "--index-url",
         "https://thirdparty.aboutcode.org/pypi/simple/",
+        "--index-url",
+        "https://pypi.org/simple/",
     ]
     check_specs_resolution(
         specifier=specifier,
@@ -161,6 +185,77 @@ def test_cli_with_multiple_index_url_and_tilde_req():
     )
 
 
+@pytest.mark.online
+def test_cli_with_single_env_var_index_url_flag_override():
+    # Click default is to override env vars via flag as shown here
+    expected_file = test_env.get_test_loc("single-url-env-var-expected.json", must_exist=False)
+    specifier = "zipp==3.8.0"
+    os.environ["PYINSP_INDEX_URL"] = "https://thirdparty.aboutcode.org/pypi/simple/"
+    extra_options = [
+        "--index-url",
+        "https://pypi.org/simple",
+    ]
+    check_specs_resolution(
+        specifier=specifier,
+        expected_file=expected_file,
+        extra_options=extra_options,
+        regen=REGEN_TEST_FIXTURES,
+    )
+    os.unsetenv("PYINSP_INDEX_URL")
+
+
+@pytest.mark.online
+def test_cli_with_single_env_var_index_url_except_pypi_simple():
+    expected_file = test_env.get_test_loc(
+        "single-url-env-var-except-simple-expected.json", must_exist=False
+    )
+    # using flask since it's not present in thirdparty
+    specifier = "flask"
+    os.environ["PYINSP_INDEX_URL"] = "https://thirdparty.aboutcode.org/pypi/simple/"
+    try:
+        check_specs_resolution(
+            specifier=specifier,
+            expected_file=expected_file,
+            extra_options=[],
+            regen=REGEN_TEST_FIXTURES,
+        )
+    except Exception as e:
+        assert "python_inspector.error.NoVersionsFound: This package does not exist: flask" in str(
+            e
+        )
+    os.unsetenv("PYINSP_INDEX_URL")
+
+
+@pytest.mark.online
+def test_cli_with_multiple_env_var_index_url_and_tilde_req():
+    expected_file = test_env.get_test_loc("tilde_req-expected-env.json", must_exist=False)
+    specifier = "zipp~=3.8.0"
+    os.environ["PYINSP_INDEX_URL"] = (
+        "https://pypi.org/simple https://thirdparty.aboutcode.org/pypi/simple/"
+    )
+    check_specs_resolution(
+        specifier=specifier,
+        expected_file=expected_file,
+        extra_options=[],
+        regen=REGEN_TEST_FIXTURES,
+    )
+    os.unsetenv("PYINSP_INDEX_URL")
+
+
+@pytest.mark.online
+def test_cli_with_single_env_var_index_url():
+    expected_file = test_env.get_test_loc("single-url-env-var-expected.json", must_exist=False)
+    specifier = "zipp==3.8.0"
+    os.environ["PYINSP_INDEX_URL"] = "https://pypi.org/simple"
+    check_specs_resolution(
+        specifier=specifier,
+        expected_file=expected_file,
+        extra_options=[],
+        regen=REGEN_TEST_FIXTURES,
+    )
+    os.unsetenv("PYINSP_INDEX_URL")
+
+
 @pytest.mark.online
 def test_cli_with_environment_marker_and_complex_ranges():
     requirements_file = test_env.get_test_loc("environment-marker-test-requirements.txt")
@@ -256,7 +351,7 @@ def test_cli_with_azure_devops_with_python_38():
 
 @pytest.mark.online
 def test_cli_with_multiple_index_url_and_tilde_req_with_max_rounds():
-    expected_file = test_env.get_test_loc("tilde_req-expected.json", must_exist=False)
+    expected_file = test_env.get_test_loc("tilde_req-expected-max-rounds.json", must_exist=False)
     specifier = "zipp~=3.8.0"
     extra_options = [
         "--index-url",
@@ -276,7 +371,7 @@ def test_cli_with_multiple_index_url_and_tilde_req_with_max_rounds():
 
 @pytest.mark.online
 def test_cli_with_multiple_index_url_and_tilde_req_and_netrc_file_without_matching_url():
-    expected_file = test_env.get_test_loc("tilde_req-expected.json", must_exist=False)
+    expected_file = test_env.get_test_loc("tilde_req-expected-netrc.json", must_exist=False)
     netrc_file = test_env.get_test_loc("test-commented.netrc", must_exist=False)
     specifier = "zipp~=3.8.0"
     extra_options = [
@@ -345,7 +440,13 @@ def test_cli_with_insecure_option():
     )
 
 
-@pytest.mark.skipif(sys.version_info[:2] == (3, 12), reason="Skipping test for Python 3.12")
+@pytest.mark.skipif(
+    sys.version_info[:2] >= (3, 12),
+    reason="Skipping test for Python 3.12+ because of "
+    "https://github.com/aboutcode-org/python-inspector/issues/212"
+    "to avoid error AttributeError: module 'configparser' has no attribute "
+    "'SafeConfigParser'. Did you mean: 'RawConfigParser'?",
+)
 @pytest.mark.online
 def test_cli_with_insecure_option_testpkh():
     setup_py_file = test_env.get_test_loc("insecure-setup-2/setup.py")
@@ -431,8 +532,8 @@ def test_passing_of_json_pdt_and_json_flags():
 
 def test_version_option():
     options = ["--version"]
-    result = run_cli(options=options)
-    assert "0.13.0" in result.output
+    rc, stdout, stderr = run_cli(options=options)
+    assert "0.14.0" in stdout
 
 
 def test_passing_of_netrc_file_that_does_not_exist():
@@ -457,29 +558,45 @@ def test_passing_of_no_json_output_flag():
 
 
 def test_passing_of_no_os():
-    options = ["--specifier", "foo", "--json", "-", "--python-version", "38"]
+    options = [
+        "--specifier",
+        "foo",
+        "--json",
+        "-",
+        "--python-version",
+        "38",
+        "--operating-system",
+        "",
+    ]
     run_cli(options=options, expected_rc=2, get_env=False)
 
 
 def test_passing_of_no_pyver():
-    options = ["--specifier", "foo", "--json", "-", "--operating-system", "linux"]
+    options = [
+        "--specifier",
+        "foo",
+        "--json",
+        "-",
+        "--operating-system",
+        "linux",
+        "--python-version",
+        "",
+    ]
     run_cli(options=options, expected_rc=2, get_env=False)
 
 
 def test_passing_of_wrong_pyver():
     options = ["--specifier", "foo", "--json", "-", "--python-version", "foo"]
     message = "Invalid value for '-p' / '--python-version'"
-    result = run_cli(options=options, expected_rc=2, get_env=False)
-    if message:
-        assert message in result.output
+    rc, stdout, stderr = run_cli(options=options, expected_rc=2, get_env=False)
+    assert message in stderr
 
 
 def test_passing_of_unsupported_os():
     options = ["--specifier", "foo", "--json", "-", "--operating-system", "bar"]
     message = "Invalid value for '-o' / '--operating-system'"
-    result = run_cli(options=options, expected_rc=2, get_env=False)
-    if message:
-        assert message in result.output
+    rc, stdout, stderr = run_cli(options=options, expected_rc=2, get_env=False)
+    assert message in stderr
 
 
 def check_requirements_resolution(
@@ -514,9 +631,9 @@ def check_setup_py_resolution(
     else:
         options = ["--setup-py", setup_py, "--json", result_file]
     options.extend(extra_options)
-    result = run_cli(options=options, expected_rc=expected_rc)
+    rc, stdout, stderr = run_cli(options=options, expected_rc=expected_rc)
     if message:
-        assert message in result.output
+        assert message in stderr
     if expected_rc == 0:
         check_json_file_results(result_file=result_file, expected_file=expected_file, regen=regen)
 
@@ -543,6 +660,7 @@ def check_data_results(results, expected_file, regen=REGEN_TEST_FIXTURES):
     results from ``results_file``. This is convenient for updating tests
     expectations.
     """
+    results = clean_results(results)
     if regen:
         with open(expected_file, "w") as exo:
             json.dump(results, exo, indent=2, separators=(",", ": "))
@@ -550,35 +668,85 @@ def check_data_results(results, expected_file, regen=REGEN_TEST_FIXTURES):
     else:
         with open(expected_file) as reso:
             expected = json.load(reso)
+
+    expected = clean_results(expected)
+
     assert results == expected
 
 
-def run_cli(options, cli=resolve_dependencies, expected_rc=0, env=None, get_env=True):
+def clean_results(results):
     """
-    Run a command line resolution. Return a click.testing.Result object.
+    Return cleaned data
     """
+    if isinstance(results, dict) and "headers" in results:
+        headers = results.get("headers", {}) or {}
+        if "tool_version" in headers:
+            del headers["tool_version"]
+        options = headers.get("options", []) or []
+        if "--verbose" in options:
+            options.remove("--verbose")
+
+    return results
+
+
+def run_cli(
+    options,
+    expected_rc=0,
+    env=None,
+    get_env=True,
+    retry=True,
+):
+    """
+    Run a python-inspector command as a plain subprocess. Return results.
+    """
+
+    from commoncode.command import execute
 
     if not env:
         env = dict(os.environ)
 
-    runner = CliRunner()
     if get_env:
         options = append_os_and_pyver_options(options)
 
     if "--generic-paths" not in options:
         options.append("--generic-paths")
 
-    result = runner.invoke(cli, options, catch_exceptions=False, env=env)
+    root_dir = dirname(dirname(__file__))
+    py_cmd = os.path.abspath(os.path.join(root_dir, "venv", "bin", "python-inspector"))
+    rc, stdout, stderr = execute(
+        cmd_loc=py_cmd,
+        args=options,
+        env=env,
+    )
 
-    if result.exit_code != expected_rc:
-        output = result.output
-        opts = " ".join(options)
+    if retry and rc != expected_rc:
+        # wait and rerun in verbose mode to get more in the output
+        time.sleep(1)
+        if "--verbose" not in options:
+            options.append("--verbose")
+        rc, stdout, stderr = execute(
+            cmd_loc=py_cmd,
+            args=options,
+            env=env,
+        )
+
+    if rc != expected_rc:
+        opts = get_opts(options)
         error = f"""
 Failure to run:
-rc: {result.exit_code}
-python-inspector {opts}
-output:
-{output}
+rc: {rc!r}
+command: python-inspector {opts}
+stdout:
+{stdout}
+
+stderr:
+{stderr}
 """
-        assert result.exit_code == expected_rc, error
-    return result
+        assert rc == expected_rc, error
+
+    return rc, stdout, stderr
+
+
+def get_opts(options):
+    opts = [o if isinstance(o, str) else repr(o) for o in options]
+    return " ".join(opts)
